/* mf - use a differential equation solver to imitate an analogue circuit */

#include "m_pd.h"
#include <math.h>
#define DIM 4
#define FLOAT double

typedef struct _params
{
    FLOAT p_input;
    FLOAT p_cutoff;
    FLOAT p_resonance;
    FLOAT p_saturation;
    FLOAT p_derivativeswere[DIM];
} t_params;

    /* imitate the (tanh) clipping function of a transistor pair.  We
    hope/assume the C compiler is smart enough to inline this so use
    a function instead of a #define. */
#if 0
static FLOAT clip(FLOAT value, FLOAT saturation, FLOAT saturationinverse)
{
    return (saturationinverse * tanh(value * saturation));
}
#else
    /* cheaper way - to 4th order, tanh is x - x*x*x/3; this cubic's
    plateaus are at +/- 1 so clip to 1 and evaluate the cubic.
    This is pretty coarse - for instance if you clip a sinusoid this way you
    can sometimes hear the discontinuity in 4th derivative at the clip point */
static FLOAT clip(FLOAT value, FLOAT saturation, FLOAT saturationinverse)
{
    float v2 = (value*saturation > 1 ? 1 :
        (value*saturation < -1 ? -1:
            value*saturation));
    return (saturationinverse * (v2 - (1./3.) * v2 * v2 * v2));
}
#endif

static void calc_derivatives(FLOAT *dstate, FLOAT *state, t_params *params)
{
    FLOAT k = 6.283*params->p_cutoff;
    FLOAT sat = params->p_saturation, satinv = 1./sat;
    dstate[0] = k *
        (clip(params->p_input - params->p_resonance * state[3], sat, satinv)
        - clip(state[0], sat, satinv));
    dstate[1] = k * (clip(state[0], sat, satinv) - clip(state[1], sat, satinv));
    dstate[2] = k * (clip(state[1], sat, satinv) - clip(state[2], sat, satinv));
    dstate[3] = k * (clip(state[2], sat, satinv) - clip(state[3], sat, satinv));
}

static void solver_euler(FLOAT *state, FLOAT *errorestimate, 
    FLOAT stepsize, t_params *params)
{
    FLOAT cumerror = 0;
    int i;
    FLOAT derivatives[DIM];
    calc_derivatives(derivatives, state, params);
    *errorestimate = 0;
    for (i = 0; i < DIM; i++)
    {
        state[i] += stepsize * derivatives[i];
        *errorestimate += (derivatives[i] > params->p_derivativeswere[i] ?
            derivatives[i] - params->p_derivativeswere[i] :
            params->p_derivativeswere[i] - derivatives[i]);
    }
    for (i = 0; i < DIM; i++)
        params->p_derivativeswere[i] = derivatives[i];
}

static void solver_rungekutte(FLOAT *state, FLOAT *errorestimate, 
    FLOAT stepsize, t_params *params)
{
    FLOAT cumerror = 0;
    int i;
    FLOAT deriv1[DIM], deriv2[DIM], deriv3[DIM], deriv4[DIM], tempstate[DIM];
    FLOAT oldstate[DIM], backstate[DIM];
#if 0
    for (i = 0; i < DIM; i++)
        oldstate[i] = state[i];
#endif
    *errorestimate = 0;
    calc_derivatives(deriv1, state, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] + 0.5 * stepsize * deriv1[i];
    calc_derivatives(deriv2, tempstate, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] + 0.5 * stepsize * deriv2[i];
    calc_derivatives(deriv3, tempstate, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] + stepsize * deriv3[i];
    calc_derivatives(deriv4, tempstate, params);
    for (i = 0; i < DIM; i++)
        state[i] += (1./6.) * stepsize * 
            (deriv1[i] + 2 * deriv2[i] + 2 * deriv3[i] + deriv4[i]);
#if 0
    calc_derivatives(deriv1, state, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] - 0.5 * stepsize * deriv1[i];
    calc_derivatives(deriv2, tempstate, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] - 0.5 * stepsize * deriv2[i];
    calc_derivatives(deriv3, tempstate, params);
    for (i = 0; i < DIM; i++)
        tempstate[i] = state[i] - stepsize * deriv3[i];
    calc_derivatives(deriv4, tempstate, params);
    for (i = 0; i < DIM; i++)
    {
        backstate[i] = state[i ]- (1./6.) * stepsize * 
            (deriv1[i] + 2 * deriv2[i] + 2 * deriv3[i] + deriv4[i]);
        *errorestimate += (backstate[i] > oldstate[i] ?
            backstate[i] - oldstate[i] : oldstate[i] - backstate[i]);
    }
#endif
}

typedef struct _mf
{
    t_object x_obj;
    t_float x_f;
    t_outlet *x_out1;    /* signal output */
    t_outlet *x_out2;    /* error estimate */
    t_params x_params;
    FLOAT x_state[DIM];
    FLOAT x_sr;
    FLOAT x_cumerror;
    int x_errorcount;
} t_mf;

static t_class *mf_class;

static void mf_resonance(t_mf *x, t_float resonance)
{
    if (resonance < 0)
        resonance = 0;
    else if (resonance > 10)
        resonance = 10;
    x->x_params.p_resonance = resonance;
}

static void mf_saturation(t_mf *x, t_float saturation)
{
    if (saturation <= 1e-3)
        saturation = 1e-3;
    x->x_params.p_saturation = saturation;
}

static void mf_clear(t_mf *x)
{
    int i;
    for (i = 0; i < DIM; i++)
        x->x_state[i] = x->x_params.p_derivativeswere[i] = 0;
}

static void mf_error(t_mf *x)
{
    outlet_float(x->x_out2,
        (x->x_errorcount ? x->x_cumerror/x->x_errorcount : 0));
    x->x_cumerror = 0;
    x->x_errorcount = 0;
}

static void mf_print(t_mf *x)
{
    int i;
    for (i = 0; i < DIM; i++)
        post("state %d: %f", i, x->x_state[i]);

}

static void *mf_new(t_float resonance)
{
    t_mf *x = (t_mf *)pd_new(mf_class);
    x->x_out1 = outlet_new(&x->x_obj, gensym("signal"));
    x->x_out2 = outlet_new(&x->x_obj, gensym("float"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
    x->x_f = 0;
    mf_clear(x);
    x->x_cumerror = 0;
    x->x_errorcount = 0;
    mf_resonance(x, resonance); 
    mf_saturation(x, 0); 
    return (x);
}

static t_int *mf_perform(t_int *w)
{
    t_mf *x = (t_mf *)(w[1]);
    t_float *in1 = (t_float *)(w[2]);
    t_float *in2 = (t_float *)(w[3]);
    t_float *out = (t_float *)(w[4]);
    int n = (int)(w[5]), i;
    FLOAT stepsize = 1/x->x_sr;
    FLOAT errorestimate;
    for (i = 0; i < n; i++)
    {
        x->x_params.p_input = *in1++;
        x->x_params.p_cutoff = *in2++;
        solver_rungekutte(x->x_state, &errorestimate, stepsize, &x->x_params);
        *out++ = x->x_state[0];
        x->x_cumerror += errorestimate;
        x->x_errorcount++;
    }
    return (w+6);
}

static void mf_dsp(t_mf *x, t_signal **sp)
{
    x->x_sr = sp[0]->s_sr;
    dsp_add(mf_perform, 5, x, sp[0]->s_vec, sp[1]->s_vec,
        sp[2]->s_vec, sp[0]->s_n);
}

void mf_tilde_setup(void)
{
    int i;
    mf_class = class_new(gensym("mf~"),
        (t_newmethod)mf_new, 0, sizeof(t_mf), 0, 0);
    class_addmethod(mf_class, (t_method)mf_saturation, gensym("saturation"),
        A_FLOAT, 0);
    class_addmethod(mf_class, (t_method)mf_resonance, gensym("resonance"),
        A_FLOAT, 0);
    class_addmethod(mf_class, (t_method)mf_clear, gensym("clear"), 0);
    class_addmethod(mf_class, (t_method)mf_print, gensym("print"), 0);
    class_addmethod(mf_class, (t_method)mf_error, gensym("error"), 0);

    class_addmethod(mf_class, (t_method)mf_dsp, gensym("dsp"), A_CANT, 0);
    CLASS_MAINSIGNALIN(mf_class, t_mf, x_f);
}
